# Copyright 2017 VMware, Inc.  All rights reserved. -- VMware Confidential
# APUAT support bundle generator.
#
# APUAT-TODO Refactoring and clean up is required

import sys
import subprocess
import os
import os.path
import errno
import time
import shutil
import logging
import getopt

try:
    VCOPS_BASE = os.environ['VCOPS_BASE']
except KeyError as ex:
    # In cloudvm, this could happen - for example, if caller like cis
    # has not called the /etc/profile.d/vcops.sh.
    filePath = os.path.dirname(os.path.realpath( __file__ ))
    # Since this file is located at $VCOPS_BASE/tools, we can use relative path
    VCOPS_BASE =  os.path.abspath(filePath + "/..")
VCOPS_BASE=VCOPS_BASE.replace('\\', '/')

commonLib = os.path.join(VCOPS_BASE ,'install')
sys.path.append(commonLib)

commonLibOld = os.path.join(VCOPS_BASE,'user','conf','install')
sys.path.append(commonLibOld)

vcopsToolsDir = os.path.join(VCOPS_BASE, 'tools')
print ('... APUAT INFO: vcopsToolsDir is: {0}'.format(vcopsToolsDir))
sys.path.append(vcopsToolsDir)

sys.path.append(os.path.join(os.path.sep,'usr', 'lib','vmware-vcopssuite','utilities','bin'))
sys.path.append(os.path.join(VCOPS_BASE, '..','vmware-vcopssuite','utilities','bin'))

from ftpuploader import *

def def_by_os(linux, windows):
    """
    choose between two values based on OS
    """
    if os.name == 'posix':
        return linux
    else:
        return windows

from installPaths import InstallPaths

## Hov-TODO - fix and refactor me for mandatory option, file existence etc.
apuatSupportBundleConfigFile = None

def usage():
    print("""usage: generateSupportBundle.py [-h] [-a ACTION] [-f FILTER] [-m] [-q]
                                [-t TIMESTAMP] [--ftp FTP]
                                [--ftp_username FTP_USERNAME]
                                [--ftp_password FTP_PASSWORD]

optional arguments:
  -h, --help            show this help message and exit
  -a ACTION             The action to take (This option is mandatory)
                        create (optional --timestamp and --filter options)
                        status (requires --timestamp option)
                        get    (get files locations,requires --timestamp)
                        delete (requires --timestamp option)
                        list
                        ftpupload (requires --timestamp --ftp --ftp_username --ftp_password)
  -f FILTER             Use along with --create to generate support
                        bundle with filtered criteria. By default,
                        it will download all configurations and log files.
                        Construct filter with following items:
                        ============Higher level filter items=============
                        1. logs
                        2. configs
                        3. sysdump
                        4. Cluster Info
                        ======== Detailed filter items for logs =========
                        1a. analytics logs
                        1b. collectors logs
                        1c. gemfire logs
                        1d. other logs
                        
                        For examples,
                        generateSupportBundle.py -a=create -m  -f=1,2,3,4
                        generateSupportBundle.py -a=create -f=1a,2,3
  -m                    Flag to get all most recent logs only, if -m is not specified, 
                        then all log files will be downloaded.
  -q                    Flag to turn on quite mode to just print message codes
  -t TIMESTAMP          timestamp for support bundle,value is long 
                         (milliseconds)
  --ftp FTP             Base URL of FTP Server, 
                        e.g., ftp://trial.vmware.com/some_path
  --ftp_username FTP_USERNAME
                        Username for FTP Server
  --ftp_password FTP_PASSWORD
                        Password for FTP Server""")

LOG_FILE = VCOPS_BASE + os.sep + "user" + os.sep + "log" + os.sep + "supportBundleGenerator.log"
SUPPORT_DIR = VCOPS_BASE + os.sep + "support"
STATUS_FILE = SUPPORT_DIR + os.sep + 'status'
LATEST_TIMESTAMP_FILE = SUPPORT_DIR + os.sep + 'timestamp'

msgDict = {'INVALID_INPUTS_CREATE_SUPPORT_BUNDLE':'-t Timestamp and -f filter are required fields..', \
           'BUSY_CREATE_SUPPORT_BUNDLE':'Another process is in progress, so exiting', \
           'PROCESS_STARTED_CREATE_SUPPORT_BUNDLE':'Process started successfully, use "-a status" to check status', \
           'STATUS_NOT_FOUND':'Tech support bundle cannot be found for timestamp: ', \
           'STATUS_SUCCESS':'Tech support bundle is created successfully for timestamp: ', \
           'STATUS_ERROR':'Tech support bundle generation is failed for timestamp: ', \
           'STATUS_PROGRESS':'Tech support bundle generation process is in progress for timestamp: '
          }

def pid_exists(pid):
    if os.name == 'posix':
        """Check whether pid exists in the current process table."""    
        if pid < 0:
            return False
        try:
            os.kill(pid, 0)
        except OSError as e:
            return e.errno == errno.EPERM
        else:
            return True
    else:
        import ctypes
        from win32com.client import GetObject
        WMI = GetObject('winmgmts:')
        processes = WMI.InstancesOf('Win32_Process')
        return pid in [process.Properties_('ProcessID').Value for process in processes]

def readlines(filename):
    try:
        f = open(filename, "r")
        try:
            lines = f.readlines()
            return lines
        finally:
            f.close()
    except IOError:
        pass
    return None

def overwritelines(filename, lines):
    try:
        f = open(filename, "w")
        try:
            f.writelines(lines)
        finally:
            f.close()
    except IOError:
        pass
    return lines


def mark_timestamp_error(timestamp, errorcodes):
    directory = SUPPORT_DIR + os.sep + timestamp
    if not os.path.exists(directory):
        os.makedirs(directory)
    filename = directory + os.sep + 'error-' + timestamp
    lines = 'errors: ' + errorcodes + '\n'
    lines = lines + 'timestamp: ' + timestamp + '\n'
    overwritelines(filename, lines)

def validate_status_file():
    if os.path.isfile(STATUS_FILE):
        lines = readlines(STATUS_FILE)
        if len(lines) < 2 or not pid_exists(int(lines[1].rstrip())):
            if len(lines) > 0:
                mark_timestamp_error(lines[0].rstrip(), "PROCESS_DIED")
            os.remove(STATUS_FILE)

def checkstatus(timestamp):
    """ If summary file is created then return STATUS_SUCCESS(success),
        if error file created then STATUS_ERROR (error),
        if status file exists and has timestamp then return STATUS_PROGRESS (pending)
        if status file exists but timestamp not found then return STATUS_NOT_FOUND(NOT_FOUND)
    """
    validate_status_file()
    if timestamp is None:
        l = readlines(LATEST_TIMESTAMP_FILE)
        if l is not None:
            timestamp = l[0].rstrip()
        else:
            return (timestamp, "STATUS_NOT_FOUND")

    if os.path.isfile(STATUS_FILE):
        lines = readlines(STATUS_FILE)
        if lines[0].rstrip() == timestamp:
            return (timestamp, "STATUS_PROGRESS")

    if os.path.isfile(SUPPORT_DIR + os.sep + timestamp + os.sep + 'summary-' + timestamp):
        return (timestamp, "STATUS_SUCCESS")
    elif os.path.isfile(SUPPORT_DIR + os.sep + timestamp + os.sep + 'error-' + timestamp):
        return (timestamp, "STATUS_ERROR")
    else:
        return (timestamp, "STATUS_NOT_FOUND")


def createSupportBundle(timestamp, filter, isMostRecent):
    if(timestamp is None or filter is None):
        return (None, 'INVALID_INPUTS_CREATE_SUPPORT_BUNDLE')
    validate_status_file()
    if os.path.isfile(STATUS_FILE):
        return (None, 'BUSY_CREATE_SUPPORT_BUNDLE')
    directory = SUPPORT_DIR + os.sep + timestamp
    if not os.path.exists(directory):
        os.makedirs(directory)
    logfile = directory + os.sep + 'log'
    java = InstallPaths.find_java_executable()
    classpath = VCOPS_BASE + "/tools/*" + def_by_os(':', ';') + \
                VCOPS_BASE + "/tomcat-enterprise/webapps/suite-api/docs/bindings/java/*" + def_by_os(':', ';') + \
                VCOPS_BASE + "/common/lib/*"
    cmd_list = [os.path.normpath(java), \
                '-Xmx1024M', '-XX:+HeapDumpOnOutOfMemoryError', \
                '-XX:HeapDumpPath=' + os.path.join(VCOPS_BASE, 'support'), \
                '-cp', classpath, \
                '-DALIVE_BASE=' + VCOPS_BASE,
                '-Dlog4j.configuration=file:' + os.path.join(VCOPS_BASE, 'tools', 'log4j.properties'), \
                'com.vcops.tools.supportbundlegen.core.SupportBundleGenerator', '-f', filter, '-t', timestamp, '-c', apuatSupportBundleConfigFile]
    if isMostRecent:
        cmd_list.append('-m')

    with open(logfile, 'w') as f1:
        processObj = subprocess.Popen(cmd_list, stdout=f1)

    pid = processObj.pid
    lines = timestamp + "\n" + str(pid) + "\n"
    overwritelines(STATUS_FILE, lines)
    overwritelines(LATEST_TIMESTAMP_FILE, timestamp)
    return (processObj, 'PROCESS_STARTED_CREATE_SUPPORT_BUNDLE')

def printfiles(timestamp, quitemode):
    timestamp, status = checkstatus(timestamp)
    if quitemode:
        print(status + " " + str(timestamp))
    else:
        print(msgDict[status] + str(timestamp))
    if status == 'STATUS_SUCCESS':
        if not quitemode:
            print ('------------Generated support bundles----------------')
        directory = SUPPORT_DIR + os.sep + timestamp
        os.chdir(directory)
        for name in os.listdir("."):
            if name.endswith(".zip"):
                print(os.path.join(directory, name))
        for name in os.listdir("."):
            if name.startswith("summary"):
                print(os.path.join(directory, name))
        for name in os.listdir("."):
            if name.startswith("log"):
                print(os.path.join(directory, name))

def uploadftp(timestamp, ftp_url, ftp_username, ftp_password):
    logging.basicConfig(filename=LOG_FILE, format='%(asctime)s %(message)s', level=logging.INFO)
    if ftp_url.index('ftp://') != 0:
        logging.error('FTP URL:' + ftp_url + ' is invalid')
        return 'FTP_URL_INVALID'
    url_split =  ftp_url[6:].rsplit('/')
    timestamp, status = checkstatus(timestamp)
    if status == 'STATUS_SUCCESS':
        logging.info('Status for timestamp:' + timestamp + " is successful, so now uploading files to FTP")
        uploader = FTPUploader()
        f = uploader.init(url_split[0], ftp_username, ftp_password, 21)
        if not f:
            return "CAN_NOT_CONNECT_TO_FTP"
        logging.info('FTP is initalized')
        success = True
        for d in url_split[1:]:
            flg1 = uploader.cwd(d)
            success = success and flg1
            logging.info('result of ftp.cwd is ' + str(flg1) + ' for dir:' + str(d))
        if not success:
            logging.error('Since changing to directory is failed, returning failure.')
            return "FTP_UPLOAD_FAILED"
        directory = SUPPORT_DIR + os.sep + timestamp
        os.chdir(directory)
        success = True
        for name in os.listdir("."):
            if name.endswith(".zip"):
                flg = uploader.upload(os.path.join(directory, name),True)
                logging.info('Upload status of file:' + os.path.join(directory, name) + " is " + str(flg))
                success = success and flg
        for name in os.listdir("."):
            if name.startswith("summary") or name.startswith("log"):
                flg = uploader.upload(os.path.join(directory, name),False)
                logging.info('Upload status of file:' + os.path.join(directory, name) + " is " + str(flg))
                success = success and flg

        logging.info('Value of success:' + str(success)) 
        if success:
            return "FTP_UPLOAD_SUCCESS"
        else:
            return "FTP_UPLOAD_FAILED"    
    else:
        return status + " " + timestamp

def sorted_ls(path):
    mtime = lambda f: os.stat(os.path.join(path, f)).st_mtime
    return list(sorted(os.listdir(path), key=mtime))

def listtimestamps():
    for name in sorted_ls(SUPPORT_DIR):
        if os.path.isdir(os.path.join(SUPPORT_DIR, name)) and checkstatus(name)[1] == 'STATUS_SUCCESS':
            print (name)

try:
    opts, args = getopt.getopt(sys.argv[1:], "ha:fc:mqt:", ["help", "action=", "filter=", "config=", "mostRecent","quitemode","timestamp=","ftp=","ftp_username=","ftp_password="])
except getopt.GetoptError as err:
    # print help information and exit:
    print (str(err)) # will print something like "option -a not recognized"
    usage()
    sys.exit(2)
output = None
verbose = False
options={"mostRecent": False, "quitemode": False}

print('APUAT BUNDLE: opts: {0}'.format(opts))
for o, a in opts:
    print('APUAT BUNDLE: o a: ({0}, {1})'.format(o, a))
    if len(a) > 0 and a[0]=="=":
        a=a[1:]
    if o in ("-h", "--help"):
        usage()
        sys.exit()
    elif o in ("-a", "--action"):
        options["action"] = a
    elif o in ("-f", "--filter"):
        options["filter"] = a
    elif o in ("-c", "--config"):
        options["config"] = a
    elif o in ("-m", "--mostRecent"):
        options["mostRecent"] = True
    elif o in ("-q", "--quitemode"):
        options["quitemode"] = True
    elif o in ("-t", "--timestamp"):
        options["timestamp"] = a
    elif o in ("--ftp"):
        options["ftp"] = a
    elif o in ("--ftp_username"):
        options["ftp_username"] = a
    elif o in ("--ftp_password"):
        options["ftp_password"] = a
    else:
        assert False, "unhandled option"

apuatSupportBundleConfigFile = options['config']

ds = []
if options.get("action") is None:
    options["action"] = 'create'
    ds.append('action   : -a=create')
if options.get("action") == 'create' and options.get("filter") is None:
    options["filter"] = "1,2,3"
    ds.append("fiter    : -f=1,2,3  (all logs,configs and support dumps; clusterInfo not included)")
if options.get("action") == 'create' and options.get("timestamp") is None:
    timestamp = int(round(time.time() * 1000))
    options["timestamp"] = str(timestamp)
    ds.append("timestamp: -t=" + str(timestamp) + " (current time in milliseconds)")
if options.get("action") == 'create' and not options.get("quitemode"):
    if len(ds) > 0:
        print ("Default Values:")
        for msg in ds:
            print (msg)
    if options.get("mostRecent"):
        print ("\nOption -m=true so only most recent logs will be bundled.")
    else:
        print ("\nOption -m=false so all log files will be downloaded, so support bundle may be large.")

if options.get("action") == 'create':
    processObj, msg = createSupportBundle(options.get("timestamp"), options.get("filter"), options.get("mostRecent"))
    if options.get("quitemode"):
        print (msg)
    else:
        if msg == 'PROCESS_STARTED_CREATE_SUPPORT_BUNDLE':
            print ("Waiting until support bundle is created\n")
            processObj.communicate()
            validate_status_file()
            t,status = checkstatus(options.get("timestamp"))
            printfiles(options.get("timestamp"), options.get("quitemode"))
        else:
            print('{0}: {1}'.format(msg, msgDict[msg]))
elif options.get("action") == 'status':
    timestamp, msg = checkstatus(options.get("timestamp"))
    if options.get("quitemode"):
        print (msg + ' ' + str(timestamp))
    else:
        if msg == 'STATUS_SUCCESS':
            printfiles(timestamp, False)
        elif timestamp is not None:
            print (msgDict[msg] + str(timestamp))
        else:
            print("No recent timestamp found.")
elif options.get("action") == 'get':
    printfiles(options.get("timestamp"), options.get("quitemode"))
elif options.get("action") == 'list':
    listtimestamps()
elif options.get("action") == 'ftpupload':
    message = uploadftp(options.get("timestamp"), options.get("ftp"), options.get("ftp_username"), options.get("ftp_password"))
    logging.info('Return value of uploadftp:' + str(message))
    print(message)
elif options.get("action") == 'delete':
    timestamp, msg = checkstatus(options.get("timestamp"))
    if (msg == 'STATUS_SUCCESS') or (msg == 'STATUS_ERROR'):
        timestamp_dir = SUPPORT_DIR + os.sep + timestamp
        if os.path.isdir(timestamp_dir):
            shutil.rmtree(timestamp_dir)
        print('SUCCESS')
    elif msg == 'STATUS_PROGRESS':
        print('STATUS_IN_PROGRESS')
    else:
        print('TIMESTAMP_NOT_FOUND')
