#!/usr/bin/python -cprint 'ERROR: You must invoke this script using $VMWARE_PYTHON_BIN'; import sys; sys.exit(5)
# Copyright 2019 VMware, Inc.  All rights reserved. -- VMware Confidential
# Common functions for the vCenter Operations Suite

__author__ = "VMware, Inc."

# imports
import io
import subprocess
import logging

import  vcops_pak_six

class NoSectionConfigParser(vcops_pak_six.configparser.ConfigParser):
   NO_SECTION = 'no-section'
   def read(self, resourceFile):
      with open(resourceFile, 'rb') as rp:
         rawFileContent = rp.read().decode('utf-8-sig')

      fileContent = vcops_pak_six.VcopsStringIO(u'[{0}]\n{1}'.format(self.NO_SECTION, rawFileContent))
      # pylint: disable=no-member
      if vcops_pak_six.VCOPS_PY3:
         self.read_file(fileContent, resourceFile)
      else:
         self.readfp(fileContent, resourceFile)
      # pylint: enable=no-member


def getProperties(property_file, property_names):
   """
    Search a property file (in java property format) for a set of properties
    :param property_names: iterable[str]: The keys of the properties to search for
    :return: dict[str]: the found values
   """

   result = {}
   if not property_names:
      return result

   try:
      if not isinstance(property_names, set):
         property_names = set(property_names)

      parser = NoSectionConfigParser()
      parser.read(property_file);
      # pylint: disable=no-member
      items = parser.items(NoSectionConfigParser.NO_SECTION)
      # pylint: enable=no-member
      for (key, value) in items:
         if key in property_names:
            result[key] = value
   except vcops_pak_six.configparser.Error as ex:
      logWarning('Failed to load properties file: {0}'.format(property_file))

   return result

def run_script(scriptWithArgs, logger, acceptableReturnCodes=None, suppressErrors=False, runningDirectory=None, waitForExit=True, useShell=True, universalNewlines=True):
   """
   Runs a script
   :param scriptWithArgs the script and arguments to run
   :param acceptableReturnCodes a list of return codes acceptable upon completion of running the script. Default is [0].
   :return tuple (boolean-script ran successfully, int-exit code, str-stderr, str-stdout)
   """
   scriptCommand = scriptWithArgs
   logger.debug('Script command: "%s"', scriptWithArgs)
   try:
      processPipe = subprocess.Popen(scriptCommand, shell=useShell, cwd=runningDirectory, stdout=subprocess.PIPE, stderr=subprocess.PIPE, universal_newlines=universalNewlines)
   except OSError as e:  # WindowsError may be thrown; can get all info from OSError
      message = 'Caught OSError executing command "%s". Exception "%s"'
      if suppressErrors:
         logger.info(message, scriptWithArgs, e)
      else:
         logger.error(message, scriptWithArgs, e)

      # If there was an errno and it was an acceptable return code, we'll still say we succeeded.
      succeeded = OSError.errno and OSError.errno in acceptableReturnCodes
      return succeeded, e.errno, e, ''

   if waitForExit:
      (processStdOut, processStdErr) = processPipe.communicate()
      logger.debug('stdout: "%s"', processStdOut)
      logger.debug('stderr: "%s', processStdErr)
      logger.debug('exit code: %d', processPipe.returncode)
      if not acceptableReturnCodes:
         acceptableReturnCodes = [0]
      if processPipe.returncode not in acceptableReturnCodes:
         message = 'Script command: "{0}" failed with exit code: {1}'.format(scriptCommand, processPipe.returncode)
         if suppressErrors:
            logger.debug(message)
         else:
            logger.error(message)
         return False, processPipe.returncode, processStdErr, processStdOut
      return True, processPipe.returncode, processStdErr, processStdOut
   else:
      logger.debug('Do not wait for script: "%s" to complete', scriptWithArgs)

